"""
Configuration management for Yaralyzer.
"""
import logging
import re
from argparse import _AppendAction, ArgumentParser, Namespace
from os import environ
from pathlib import Path
from typing import Any, Callable, TypeVar

from yaralyzer.output.theme import YARALYZER_THEME_DICT
from yaralyzer.util.classproperty import classproperty
from yaralyzer.util.constants import KILOBYTE, YARALYZE, YARALYZER_UPPER
from yaralyzer.util.helpers.collections_helper import listify
from yaralyzer.util.helpers.debug_helper import print_stack
from yaralyzer.util.helpers.env_helper import (is_env_var_set_and_not_false,
     is_invoked_by_pytest, is_path_var, load_dotenv_file, stderr_console, temporary_argv)
from yaralyzer.util.helpers.string_helper import is_falsey, is_number, is_truthy, log_level_for

LOG_DIR_ENV_VAR = "LOG_DIR"
LOG_LEVEL_ENV_VAR = "LOG_LEVEL"
T = TypeVar('T')

# For when we need to build a default config outside of CLI usage. TODO: kinda janky
DEFAULT_ARGV = [YARALYZE, __file__, '--regex-pattern', 'foobar']

# These options cannot be read from an environment variable
ONLY_CLI_ARGS = [
    'env_vars',
    'extract_binary_streams',
    'file_to_scan_path',
    'help',
    'interact',
    'version',
]


class YaralyzerConfig:
    """Handles parsing of command line args and environment variables for Yaralyzer."""

    ENV_VAR_PREFIX = YARALYZER_UPPER     # Yaralyzer env vars are always prefixed with this
    COLOR_THEME = YARALYZER_THEME_DICT   # Overloaded in pdfalyzer

    # These are passed through to `yara.set_config()``.
    DEFAULT_MAX_MATCH_LENGTH = 100 * KILOBYTE
    DEFAULT_YARA_STACK_SIZE = 2 * 65536
    # Skip decoding binary matches under/over these lengths
    DEFAULT_MIN_DECODE_LENGTH = 1
    DEFAULT_MAX_DECODE_LENGTH = 256
    # chardet.detect() related
    DEFAULT_MIN_CHARDET_TABLE_CONFIDENCE = 2
    DEFAULT_MIN_CHARDET_BYTES = 9
    # Number of bytes to show before/after byte previews and decodes. Configured by command line or env var
    DEFAULT_SURROUNDING_BYTES = 64
    # EncodingDetector defaults
    DEFAULT_FORCE_DISPLAY_THRESHOLD = 20.0
    DEFAULT_FORCE_DECODE_THRESHOLD = 50.0

    # Logging stuff
    LOG_DIR: Path | None = None
    LOG_LEVEL: int = logging.WARNING

    _append_option_vars: list[str] = []
    _argparse_dests: list[str] = []
    _parse_arguments: Callable[[type['YaralyzerConfig'], Namespace | None], Namespace]

    @classproperty
    def app_name(cls) -> str:
        return cls.ENV_VAR_PREFIX.strip('_').title()

    @classproperty
    def args(cls) -> Namespace:
        """Source for parsed command line arguments merged with environment variable options."""
        if '_args' not in dir(cls):
            cls._set_default_args()

        return cls._args

    @classproperty
    def executable_name(cls) -> str:
        """The command used to run this app, e.g. `'yaralyze'`."""
        return cls.app_name.lower().removesuffix('r')

    @classproperty
    def log_dir_env_var(cls) -> str:
        """Environment variable name that can set the log output directory."""
        return cls.prefixed_env_var(LOG_DIR_ENV_VAR)

    @classmethod
    def init(
        cls,
        argparser: ArgumentParser,
        parse_arguments: Callable[[type['YaralyzerConfig'], Namespace | None], Namespace]
    ) -> None:
        """
        Should be called immediately upon package load.

        Args:
            argparser (ArgumentParser): An ArgumentParser that can parse the args this app needs.
            parse_arguments (Callable): Function that can fill in and error check what `argparser.parse_args()` returns.
        """
        cls._set_class_vars_from_env()
        cls._argument_parser = argparser
        cls._argparse_dests = sorted([action.dest for action in argparser._actions])
        cls._append_option_vars = [a.dest for a in argparser._actions if isinstance(a, _AppendAction)]
        cls._parse_arguments = parse_arguments

    @classmethod
    def env_var_for_command_line_option(cls, option: str) -> str:
        """`output_dir' becomes``YARALYZER_OUTPUT_DIR`. Overriden in pdfalyzer to distinguish yaralyzer only options."""
        return cls.prefixed_env_var(option)

    @classmethod
    def get_env_value(cls, var: str, var_type: Callable[[str], T] = str) -> T | None:
        """If called with `'output_dir'` it will check env value of `YARALYZER_OUTPUT_DIR`."""
        env_var = cls.env_var_for_command_line_option(var)
        env_value = environ.get(env_var)
        var = var.removeprefix(f"{cls.ENV_VAR_PREFIX}_").lower()  # Accomodates being called with YARALYZER_OUTPUT_DIR

        # Override type for a few important situations
        if not env_value:
            return None
        elif var.lower() in cls._append_option_vars:
            env_value = env_value.split(',')
        elif is_number(env_value):
            env_value = float(env_value) if '.' in env_value else int(env_value)
        else:
            env_value = var_type(env_value)

        if is_path_var(var):
            env_value = [Path(p) for p in env_value] if isinstance(env_value, list) else Path(env_value)

            for file_path in listify(env_value):
                if not file_path.exists():
                    raise EnvironmentError(f"Environment has {env_var} set to '{env_value}' but that path doesn't exist!")

        # print(f"Got value for var='{var}', env_var='{env_var}', value={env_value}")
        if is_falsey(str(env_value)):
            return False
        elif is_truthy(str(env_value)):
            return True
        else:
            return env_value

    @classmethod
    def parse_args(cls) -> Namespace:
        """Parse `sys.argv` and merge the result with any options set in the environment variables."""
        args = cls._parse_arguments(cls, None)
        cls._merge_env_options(args)
        return args

    @classmethod
    def prefixed_env_var(cls, var: str) -> str:
        """Turns 'LOG_DIR' into 'YARALYZER_LOG_DIR' etc."""
        return (var if var.startswith(cls.ENV_VAR_PREFIX) else f"{cls.ENV_VAR_PREFIX}_{var}").upper()

    @classmethod
    def _get_default_arg(cls, arg: str) -> Any:
        """Return the default value for `arg` as defined by a `DEFAULT_` style class variable."""
        return vars(cls).get(f"DEFAULT_{arg.upper()}")

    @classmethod
    def _is_configurable_by_env_var(cls, option: str) -> bool:
        """Returns `True` if this option can be configured by a `YARALYZER_VAR_NAME` style environment variable."""
        return not (option.startswith('export') or option in ONLY_CLI_ARGS)

    @classmethod
    def _merge_env_options(cls, _args: Namespace) -> None:
        """
        Set the `args` class instance variable and update args with any environment variable overrides.
        For each arg the environment will be checked for a variable with the same name, uppercased and
        prefixed by "YARALYZER_".

        Example:
            For the argument `--output-dir` the environment will be checked for `YARALYZER_OUTPUT_DIR`.

        Args:
            _args (Namespace): Object returned by `ArgumentParser.parse_args()`
        """
        cls._args = _args

        for option in cls._argparse_dests:
            arg_value = vars(_args).get(option)
            env_value = cls.get_env_value(option)
            default_value = cls._get_default_arg(option)
            # print(f"option: {option}, arg_value: {arg_value}, env_var: {env_var}, env_value: {env_value}, default: {default_value}", file=stderr)  # noqa: E501

            # TODO: as is you can't override env vars with CLI args
            if isinstance(arg_value, bool):
                env_var = cls.prefixed_env_var(option)
                setattr(_args, option, arg_value or is_env_var_set_and_not_false(env_var))
            elif isinstance(arg_value, (int, float)):
                # Check against defaults to avoid overriding env var configured options
                if arg_value == default_value and env_value is not None:
                    setattr(_args, option, type(arg_value)(env_value) or arg_value)
            elif arg_value in ['', [], {}]:
                setattr(_args, option, env_value if env_value else arg_value)
            else:
                setattr(_args, option, arg_value or env_value)

        cls.args.output_dir = (cls.args.output_dir or Path.cwd()).resolve()
        cls.args.file_prefix = (cls.args.file_prefix + '__') if cls.args.file_prefix else ''
        cls.args.file_suffix = ('_' + cls.args.file_suffix) if cls.args.file_suffix else ''

    @classmethod
    def _set_class_vars_from_env(cls) -> None:
        """Check the environment for LOG_LEVEL and LOG_DIR so the log setter upper can use them."""
        load_dotenv_file(cls.app_name.lower())

        if (log_dir := cls.get_env_value(LOG_DIR_ENV_VAR, Path)):
            cls.LOG_DIR = Path(log_dir).resolve()

        if (log_level := cls.get_env_value(LOG_LEVEL_ENV_VAR)):
            cls.LOG_LEVEL = log_level_for(log_level)

        if cls.LOG_DIR and not is_invoked_by_pytest():
            stderr_console.print(f"Writing logs to '{cls.LOG_DIR}' instead of stderr/stdout...", style='dim')

    @classmethod
    def _set_default_args(cls) -> None:
        """Set `self.args` to their defaults as if parsed from the command line."""
        from yaralyzer.util.logging import log
        log.warning(f"{type(cls).__name__}._set_default_args() called which shouldn't be happening any more.")
        # print_stack()

        with temporary_argv(DEFAULT_ARGV):
            cls._merge_env_options(cls._parse_arguments(cls, None))



"""
Handle logging for `yaralyzer`.

There's two possible log sinks other than `STDOUT`:

  1. 'log' - the application log (standard log, what goes to `STDOUT` with `-D` option)
  2. 'invocation_log' - tracks the exact command yaralyzer was invoked with, similar to a history file

The regular log file at `APPLICATION_LOG_PATH` is where the quite verbose application logs
will be written if things ever need to get that formal. For now those logs are only accessible
on `STDOUT` with the `-D` flag but the infrastructure for persistent logging exists if someone
needs/wants that sort of thing.

Logs are not normally ephemeral/not written to files but can be configured to do so by setting
the `YARALYZER_LOG_DIR` env var. See `.yaralyzer.example` for documentation about the side effects
of setting `YARALYZER_LOG_DIR` to a value.

* [logging.basicConfig](https://docs.python.org/3/library/logging.html#logging.basicConfig)

* [realpython.com/python-logging/](https://realpython.com/python-logging/)

Python log levels for reference:

```
    CRITICAL 50
    ERROR 40
    WARNING 30
    INFO 20
    DEBUG 10
    NOTSET 0
```
"""
import logging
import time
from argparse import Namespace
from contextlib import contextmanager
from copy import copy
from pathlib import Path
from sys import argv
from typing import Any, Generator

from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.logging import RichHandler
from rich.text import Text

from yaralyzer.config import YaralyzerConfig
from yaralyzer.output.theme import LOG_THEME
from yaralyzer.util.constants import ECHO_COMMAND_OPTION, YARALYZER
from yaralyzer.util.helpers.env_helper import default_console_kwargs, is_invoked_by_pytest
from yaralyzer.util.helpers.file_helper import file_size_str, relative_path
from yaralyzer.util.helpers.string_helper import log_level_for

ARGPARSE_LOG_FORMAT = '{0: >29}    {1: <11} {2: <}\n'
LOG_FILE_LOG_FORMAT = '%(asctime)s %(levelname)s %(message)s'
LOG_SEPARATOR = '-' * 35
WRITE_STYLE = 'grey46'

DEFAULT_LOG_HANDLER_KWARGS = {
    'console': Console(stderr=True, theme=LOG_THEME, **default_console_kwargs()),
    'omit_repeated_times': False,
    'rich_tracebacks': True,
    'show_path': not is_invoked_by_pytest(),
    'show_time': not is_invoked_by_pytest(),
}


def configure_logger(log_label: str) -> logging.Logger:
    """
    Set up a file or stream `logger` depending on the configuration.

    Args:
        log_label (str): The label for the `logger`, e.g. "run" or "invocation".
            Actual name will be `"yaralyzer.{log_label}"`.

    Returns:
        logging.Logger: The configured `logger`.
    """
    log_name = f"{YARALYZER}.{log_label}"
    logger = logging.getLogger(log_name)
    rich_stream_handler = RichHandler(**DEFAULT_LOG_HANDLER_KWARGS)

    if YaralyzerConfig.LOG_DIR:
        if not (YaralyzerConfig.LOG_DIR.is_dir() and YaralyzerConfig.LOG_DIR.is_absolute()):
            raise FileNotFoundError(f"Log dir '{YaralyzerConfig.LOG_DIR}' doesn't exist or is not absolute")

        log_file_path = YaralyzerConfig.LOG_DIR.joinpath(f"{log_name}.log")
        log_file_handler = logging.FileHandler(log_file_path)
        log_file_handler.setFormatter(logging.Formatter(LOG_FILE_LOG_FORMAT))
        logger.addHandler(log_file_handler)
        rich_stream_handler.setLevel('WARN') # Rich handler is only for warnings when writing to log file

    logger.addHandler(rich_stream_handler)
    logger.setLevel(YaralyzerConfig.LOG_LEVEL)
    return logger


def invocation_str(_argv: list[str] | None = None, raw: bool = False) -> str:
    """Convert `sys.argv` into something readable by relativizing paths."""
    _argv = copy(_argv or argv)
    _argv = [arg for arg in _argv if arg != ECHO_COMMAND_OPTION]

    if not raw:
        _argv = [Path(_argv[0]).name] + [a if a.startswith('-') else str(relative_path(a)) for a in _argv[1:]]

    return "   " + ' '.join(_argv)


def invocation_txt() -> Text:
    txt = Text(f"Invoked with this command:\n\n")
    txt.append(f"{invocation_str()}\n\n", style='wheat4')

    # TODO: Ugly way to keep local system info out of fixture data
    if not is_invoked_by_pytest():
        txt.append(f"Invocation raw argv:\n\n", style='dim')
        txt.append(f"{invocation_str(raw=True)}", style='wheat4 dim')

    return txt


def log_and_print(msg: str, log_level: str = 'INFO', style: str = '') -> None:
    """Both print (to console) and log (to file) a string."""
    log.log(logging.getLevelName(log_level), msg)
    log_console.print(msg, style=style)


def log_argparse_result(args: Namespace, label: str) -> None:
    """Logs the result of `argparse`."""
    args_dict = vars(args)
    log_msg = f'{label} argparse results:\n\n' + ARGPARSE_LOG_FORMAT.format('OPTION', 'TYPE', 'VALUE')
    log_msg += f"{ARGPARSE_LOG_FORMAT.format('------', '----', '-----')}"

    for arg_var in sorted(args_dict.keys()):
        arg_val = args_dict[arg_var]
        row = ARGPARSE_LOG_FORMAT.format(arg_var, type(arg_val).__name__, str(arg_val))
        log_msg += row

    log_msg += "\n"
    invocation_log.debug(log_msg)
    log.debug(log_msg)


def log_bigly(msg: str, big_msg: object, level: int = logging.INFO) -> None:
    """Log something with newlines around it."""
    log.log(level, f"{msg}\n\n {big_msg}\n")


def log_current_config(config: type[YaralyzerConfig]) -> None:
    """Write current state of `YaralyzerConfig` object (including parsed args) to the logs."""
    msg = f"{config.__name__} current attributes:\n\n"

    config_dict = {
        k: v for k, v in vars(config).items()
        if not (k.startswith('_') or 'classmethod' in str(v))
    }

    for k in sorted(config_dict.keys()):
        msg += f"   {k: >35}  {config_dict[k]}\n"

    log.info(msg)
    log_argparse_result(config.args, 'Parsed with env vars merged')


@contextmanager
def log_file_export(file_path: Path) -> Generator[Any, Any, Any]:
    """Standardize the way file exports are logged about."""
    if file_path.exists():
        log.debug(f"Overwriting existing '{file_path}' ({file_size_str(file_path)})...")
        file_path.unlink()

    started_at = time.perf_counter()
    yield file_path
    write_time = f"{time.perf_counter() - started_at:.3f} seconds"

    if file_path.exists():
        size = file_size_str(file_path)
        log_and_print(f"Wrote '{relative_path(file_path)}' in {write_time} ({size}).", style=WRITE_STYLE)
    else:
        log.error(f"Spent {write_time} writing file '{file_path}' but there's nothing there!")


def log_trace(*args) -> None:
    """Log below logging.DEBUG level."""
    log.log(logging.NOTSET, *args)


def set_log_level(level: str | int) -> None:
    """Set the log level at any time."""
    for handler in log.handlers + [log]:
        handler.setLevel(log_level_for(level))


# See file comment. 'log' is the standard application log, 'invocation_log' is a history of yaralyzer runs
log_console = DEFAULT_LOG_HANDLER_KWARGS['console']
log = configure_logger('run')
invocation_log = configure_logger('invocation')
highlighter = ReprHighlighter()

# If we're logging to files make sure invocation_log has the right level
if YaralyzerConfig.LOG_DIR:
    invocation_log.setLevel('INFO')

# Suppress annoying chardet library logs
for submodule in ['universaldetector', 'charsetprober', 'codingstatemachine']:
    logging.getLogger(f"chardet.{submodule}").setLevel(logging.WARNING)



"""
`EncodingDetector` class for managing chardet encoding detection.
"""
from dataclasses import dataclass, field
from operator import attrgetter
from typing import ClassVar, List

import chardet
from chardet.resultdict import ResultDict
from rich import box
from rich.padding import Padding
from rich.table import Table

from yaralyzer.config import YaralyzerConfig
from yaralyzer.encoding_detection.encoding_assessment import ENCODING, EncodingAssessment
from yaralyzer.util.logging import log

CONFIDENCE_SCORE_RANGE = range(0, 101)


@dataclass
class EncodingDetector:
    """
    Manager class to ease dealing with the encoding detection library `chardet`.

    Each instance of this class manages a `chardet.detect_all()` scan on a single set of bytes.

    Attributes:
        bytes (bytes): The bytes to analyze.
        bytes_len (int): The length of the bytes.
        table (Table): A rich `Table` object summarizing the chardet results.
        assessments (List[EncodingAssessment]): List of `EncodingAssessment` objects from `chardet` results.
        unique_assessments (List[EncodingAssessment]): Unique assessments by encoding, highest confidence only.
        raw_chardet_assessments (List[dict]): Raw list of dicts returned by `chardet.detect_all()`.
        force_decode_assessments (List[EncodingAssessment]): Assessments above force decode threshold.
        force_display_assessments (List[EncodingAssessment]): Assessments above force display threshold.
        has_any_idea (bool | None): `True` if `chardet` had any idea what the encoding might be,
            `False` if not, `None` if `chardet` wasn't run yet.
        force_display_threshold (float): `[class variable]` Default confidence threshold for forcing display
            in decoded table.
        force_decode_threshold (float): `[class variable]` Default confidence threshold for forcing a decode attempt.
    """

    _bytes: bytes
    assessments: list[EncodingAssessment] = field(default_factory=list)
    force_decode_assessments: list[EncodingAssessment] = field(default_factory=list)
    force_display_assessments: list[EncodingAssessment] = field(default_factory=list)
    has_any_idea: bool | None = None
    raw_chardet_assessments: list[ResultDict] = field(default_factory=list)
    table: Table = field(default_factory=lambda: _empty_chardet_results_table())
    unique_assessments: list[EncodingAssessment] = field(default_factory=list)

    @property
    def bytes(self) -> bytes:
        return self._bytes

    @property
    def bytes_len(self) -> int:
        return len(self.bytes)

    def __post_init__(self) -> None:
        # Skip chardet if there's not enough bytes available
        if not self.has_enough_bytes():
            log.debug(f"{self.bytes_len} is not enough bytes to run chardet.detect()")
            return

        # Unique by encoding, ignoring language.  Ordered from highest to lowest confidence
        self.raw_chardet_assessments = chardet.detect_all(self.bytes, ignore_threshold=True)

        if len(self.raw_chardet_assessments) == 1 and self.raw_chardet_assessments[0][ENCODING] is None:
            log.info(f"chardet.detect() has no idea what the encoding is, result: {self.raw_chardet_assessments}")
            self.has_any_idea = False
            return

        self.has_any_idea = True
        self.assessments = [EncodingAssessment(a) for a in self.raw_chardet_assessments]
        self._uniquify_results_and_build_table()
        self.force_decode_assessments = self.assessments_above_confidence(YaralyzerConfig.args.force_decode_threshold)
        self.force_display_assessments = self.assessments_above_confidence(YaralyzerConfig.args.force_display_threshold)

    def get_encoding_assessment(self, encoding: str) -> EncodingAssessment:
        """
        Get the `chardet` assessment for a specific encoding.

        Args:
            encoding (str): The encoding to look for.

        Returns:
            EncodingAssessment: Assessment for the given encoding if it exists, otherwise a dummy with 0 confidence.
        """
        assessment = next((r for r in self.unique_assessments if r.encoding == encoding), None)
        return assessment or EncodingAssessment.dummy_encoding_assessment(encoding)

    def has_enough_bytes(self) -> bool:
        """Return `True` if we have enough bytes to run `chardet.detect()`."""
        return self.bytes_len >= YaralyzerConfig.args.min_chardet_bytes

    def assessments_above_confidence(self, cutoff: float) -> List[EncodingAssessment]:
        """Return the assessments above the given confidence cutoff."""
        return [a for a in self.unique_assessments if a.confidence >= cutoff]

    def __rich__(self) -> Padding:
        return Padding(self.table, (0, 0, 0, 0))

    def _uniquify_results_and_build_table(self) -> None:
        """Keep the highest result per encoding, ignoring the language `chardet` has indicated."""
        already_seen_encodings = {}

        for i, result in enumerate(self.assessments):
            if result.confidence < YaralyzerConfig.args.min_chardet_table_confidence:
                continue

            self.table.add_row(f"{i + 1}", result.encoding_label, result.confidence_text)

            # self.unique_assessments retains one result per encoding possibility (the highest confidence one)
            # Some encodings are not language specific and for those we don't care about the language
            if result.encoding not in already_seen_encodings:
                self.unique_assessments.append(result)
                already_seen_encodings[result.encoding] = result
            else:
                log.debug(f"Skipping chardet result {result} (already saw {already_seen_encodings[result.encoding]})")

        self.unique_assessments.sort(key=attrgetter('confidence'), reverse=True)


def _empty_chardet_results_table() -> Table:
    """Returns an empty `Table` with appropriate columns for `chardet` results."""
    table = Table(
        'Rank', 'Encoding', 'Confidence',
        title='chardet.detect results',
        title_style='color(153) italic dim',
        header_style='off_white',
        style='dim',
        box=box.SIMPLE,
        show_edge=False,
        collapse_padding=True
    )

    table.columns[0].justify = 'right'
    return table



"""
Handling of internal YARA errors.
"""
import re

import yara

INTERNAL_ERROR_REGEX = re.compile(r"internal error: (\d+)$")
YARA_ERRORS_REPO_PATH = 'master/libyara/include/yara/error.h'
YARA_ERRORS_RAW_URL = f"https://raw.githubusercontent.com/VirusTotal/yara/refs/heads/{YARA_ERRORS_REPO_PATH}"
YARA_ERRORS_URL = f"https://github.com/VirusTotal/yara/blob/{YARA_ERRORS_REPO_PATH}"

# Extracted from YARA_ERRORS_RAW_URL
YARA_ERROR_CODES = {
    0: 'SUCCESS',
    1: 'INSUFICIENT_MEMORY',
    1: 'INSUFFICIENT_MEMORY',
    2: 'COULD_NOT_ATTACH_TO_PROCESS',
    3: 'COULD_NOT_OPEN_FILE',
    4: 'COULD_NOT_MAP_FILE',
    6: 'INVALID_FILE',
    7: 'CORRUPT_FILE',
    8: 'UNSUPPORTED_FILE_VERSION',
    9: 'INVALID_REGULAR_EXPRESSION',
    10: 'INVALID_HEX_STRING',
    11: 'SYNTAX_ERROR',
    12: 'LOOP_NESTING_LIMIT_EXCEEDED',
    13: 'DUPLICATED_LOOP_IDENTIFIER',
    14: 'DUPLICATED_IDENTIFIER',
    15: 'DUPLICATED_TAG_IDENTIFIER',
    16: 'DUPLICATED_META_IDENTIFIER',
    17: 'DUPLICATED_STRING_IDENTIFIER',
    18: 'UNREFERENCED_STRING',
    19: 'UNDEFINED_STRING',
    20: 'UNDEFINED_IDENTIFIER',
    21: 'MISPLACED_ANONYMOUS_STRING',
    22: 'INCLUDES_CIRCULAR_REFERENCE',
    23: 'INCLUDE_DEPTH_EXCEEDED',
    24: 'WRONG_TYPE',
    25: 'EXEC_STACK_OVERFLOW',
    26: 'SCAN_TIMEOUT',
    27: 'TOO_MANY_SCAN_THREADS',
    28: 'CALLBACK_ERROR',
    29: 'INVALID_ARGUMENT',
    30: 'TOO_MANY_MATCHES',
    31: 'INTERNAL_FATAL_ERROR',
    32: 'NESTED_FOR_OF_LOOP',
    33: 'INVALID_FIELD_NAME',
    34: 'UNKNOWN_MODULE',
    35: 'NOT_A_STRUCTURE',
    36: 'NOT_INDEXABLE',
    37: 'NOT_A_FUNCTION',
    38: 'INVALID_FORMAT',
    39: 'TOO_MANY_ARGUMENTS',
    40: 'WRONG_ARGUMENTS',
    41: 'WRONG_RETURN_TYPE',
    42: 'DUPLICATED_STRUCTURE_MEMBER',
    43: 'EMPTY_STRING',
    44: 'DIVISION_BY_ZERO',
    45: 'REGULAR_EXPRESSION_TOO_LARGE',
    46: 'TOO_MANY_RE_FIBERS',
    47: 'COULD_NOT_READ_PROCESS_MEMORY',
    48: 'INVALID_EXTERNAL_VARIABLE_TYPE',
    49: 'REGULAR_EXPRESSION_TOO_COMPLEX',
    50: 'INVALID_MODULE_NAME',
    51: 'TOO_MANY_STRINGS',
    52: 'INTEGER_OVERFLOW',
    53: 'CALLBACK_REQUIRED',
    54: 'INVALID_OPERAND',
    55: 'COULD_NOT_READ_FILE',
    56: 'DUPLICATED_EXTERNAL_VARIABLE',
    57: 'INVALID_MODULE_DATA',
    58: 'WRITING_FILE',
    59: 'INVALID_MODIFIER',
    60: 'DUPLICATED_MODIFIER',
    61: 'BLOCK_NOT_READY',
    62: 'INVALID_PERCENTAGE',
    63: 'IDENTIFIER_MATCHES_WILDCARD',
    64: 'INVALID_VALUE',
    65: 'TOO_SLOW_SCANNING',
    66: 'UNKNOWN_ESCAPE_SEQUENCE',
}


def yara_error_msg(exception: yara.Error) -> str:
    """Turn a mysterious YARA error code number into a human readable string."""
    if (internal_error_match := INTERNAL_ERROR_REGEX.search(str(exception))):
        error_code = int(internal_error_match.group(1))
        msg = f"Internal YARA error! (code: {error_code}, type: {YARA_ERROR_CODES[error_code]})"

        if error_code == 25:
            msg += "\n\nRunning with a larger --yara-stack-size may solve this problem."

        return msg
    else:
        return f"YARA error: {exception}"



"""
Helper methods to work with files.
"""
from datetime import datetime
from pathlib import Path

from yaralyzer.util.constants import KILOBYTE, MEGABYTE


def file_size(file_path: Path | str) -> int:
    return Path(file_path).stat().st_size


def file_size_str(file_path, digits: int | None = None):
    return file_size_to_str(file_size(file_path), digits)


def file_size_to_str(size: int, digits: int | None = None) -> str:
    _digits = 2

    if size > MEGABYTE:
        size_num = float(size) / MEGABYTE
        size_str = 'MB'
    elif size > KILOBYTE:
        size_num = float(size) / KILOBYTE
        size_str = 'kb'
        _digits = 1
    else:
        return f"{size} b"

    digits = _digits if digits is None else digits
    return f"{size_num:,.{digits}f} {size_str}"


def files_in_dir(_dir: Path | str, with_extname: str | None = None) -> list[Path]:
    """
    Returns paths for all non dot files in `dir` (optionally filtered to only those ending in 'with_extname').

    Args:
        dir (Path | str): Directory to list files from.
        with_extname (str | None): If set, only return files with this extension. Defaults to None.

    Returns:
        list[Path]: List of file paths.
    """
    dir = Path(_dir)
    with_extname = f".{with_extname}" if (with_extname and not with_extname.startswith('.')) else ''
    glob_pattern = f"*{with_extname}"

    if not dir.is_dir():
        raise FileNotFoundError(f"'{_dir}' is not a directory!")

    return [f for f in dir.glob(glob_pattern) if not (f.name.startswith('.') or f.is_dir())]


def load_binary_data(file_path: Path | str) -> bytes:
    """Load and return the raw `bytes` from a file."""
    return Path(file_path).read_bytes()


def load_file(file_path: Path | str) -> str:
    """Load and return the text contents of a file."""
    return Path(file_path).read_text(encoding='utf-8')  # Windows requires forcing the encoding


def relative_path(path: Path | str) -> Path:
    """Get path relative to current working directory."""
    try:
        return Path(path).relative_to(Path.cwd())
    except ValueError:
        return Path(path)


def timestamp_for_filename() -> str:
    """Returns a string showing current time in a file name friendly format."""
    return datetime.now().strftime("%Y-%m-%dT%H.%M.%S")


def to_paths(files: list[str] | list[Path] | list[str | Path]) -> list[Path]:
    return [Path(f) for f in files]



"""
Methods to build the `rich.table` used to display decoding attempts of a given bytes array.

Final output should be a `rich.table` of decoding attempts that are sorted like this:

    1. String representation of undecoded bytes is always the first row

    2. Encodings which `chardet.detect()` ranked as > 0% likelihood are sorted based on that confidence

    3. Then the unchardetectable:

        1. Decodings that were successful, unforced, and new

        2. Decodings that were "successful" but forced

        3. Decodings that were the same as other decodings

        4. Failed decodings
"""
from rich import box
from rich.table import Table
from rich.text import Text

from yaralyzer.bytes_match import BytesMatch
from yaralyzer.util.helpers.bytes_helper import ascii_view_of_raw_bytes, hex_view_of_raw_bytes, rich_text_view_of_raw_bytes
from yaralyzer.util.helpers.rich_helper import DEFAULT_TABLE_OPTIONS, na_txt

HEX = Text('HEX', style='bytes.title')
RAW_BYTES = Text('Raw', style=f"bytes")


def new_decoding_attempts_table(bytes_match: BytesMatch) -> Table:
    """Build a new rich `Table` with two rows, the raw and hex views of the `bytes_match` data."""
    table = Table(show_lines=True, border_style='bytes', header_style='decode.table_header', **DEFAULT_TABLE_OPTIONS)

    def add_col(title, **kwargs):
        kwargs['justify'] = kwargs.get('justify', 'center')
        table.add_column(title, overflow='fold', vertical='middle', **kwargs)

    add_col('Encoding', justify='right', width=12)
    add_col('Detect Odds', width=len('Detect'))
    add_col('Used\nForce?', width=len('Force?'))
    add_col('Decoded Output', justify='left')

    na = na_txt(style=HEX.style)
    table.add_row(HEX, na, na, _hex_preview_subtable(bytes_match))
    na = na_txt(style=RAW_BYTES.style)
    table.add_row(RAW_BYTES, na, na, rich_text_view_of_raw_bytes(bytes_match.surrounding_bytes, bytes_match))
    return table


def _hex_preview_subtable(bytes_match: BytesMatch) -> Table:
    """
    Build a sub `Table` for hex view row (hex on one side, ascii on the other side).

    Args:
        bytes_match (BytesMatch): The `BytesMatch` object containing the bytes to display.

    Returns:
        Table: A `rich.table` with hex and ascii views of the bytes.
    """
    hex_table = Table(
        'hex',
        'ascii',
        border_style='grey.darkest',
        header_style='decode.table_header',
        box=box.MINIMAL,
        show_lines=True,
        show_header=True,
        show_edge=False,
        padding=(0, 1, 0, 2),
        pad_edge=False
    )

    hex_table.add_row(
        hex_view_of_raw_bytes(bytes_match.surrounding_bytes, bytes_match),
        ascii_view_of_raw_bytes(bytes_match.surrounding_bytes, bytes_match)
    )

    return hex_table



"""
Methods for computing and displaying various file hashes.
"""
import hashlib
from dataclasses import asdict, dataclass
from pathlib import Path
# from typing import Self  # TODO: this requires python 3.11

from rich.console import JustifyMethod
from rich.table import Column, Table

from yaralyzer.output.theme import GREY
from yaralyzer.util.helpers.rich_helper import DEFAULT_TABLE_OPTIONS, size_text


@dataclass
class BytesInfo:
    """Compute the size, MD5, SHA1, and SHA256 hashes for some bytes."""

    _bytes: bytes

    @property
    def size(self) -> int:
        return len(self._bytes)

    @property
    def md5(self) -> str:
        return hashlib.md5(self._bytes).hexdigest().upper()

    @property
    def sha1(self) -> str:
        return hashlib.sha1(self._bytes).hexdigest().upper()

    @property
    def sha256(self) -> str:
        return hashlib.sha256(self._bytes).hexdigest().upper()

    def hash_dict(self) -> dict[str, str]:
        return {
            'md5': self.md5,
            'sha1': self.sha1,
            'sha256': self.sha256,
        }

    @classmethod
    def for_file(cls, file_path: str | Path) -> 'BytesInfo':
        """Alternate constructor that reads the bytes from `file_path`."""
        with open(file_path, 'rb') as file:
            return cls(Path(file_path).read_bytes())


def bytes_hashes_table(
    bytes_or_info: bytes | BytesInfo,
    title: str | None = None,
    title_justify: JustifyMethod = 'left'
) -> Table:
    """
    Build a Rich `Table` displaying the size, MD5, SHA1, and SHA256 hashes of a byte sequence.

    Args:
        bytes_or_info (Union[bytes, BytesInfo]): The `bytes` to hash, or a `BytesInfo`
            namedtuple with precomputed values.
        title (str | None, optional): Optional title for the table. Defaults to `None`.
        title_justify (JustifyMethod, optional): Justification for the table title. Defaults to `"LEFT"`.

    Returns:
        Table: A Rich `Table` object with the size and hash values.
    """
    bytes_info = bytes_or_info if isinstance(bytes_or_info, BytesInfo) else BytesInfo(bytes_or_info)

    table = Table(
        'Size',
        Column(size_text(bytes_info.size)),
        title=f" {title} Bytes Info" if title else None,
        title_style=GREY,
        title_justify=title_justify,
        **DEFAULT_TABLE_OPTIONS,
    )

    table.columns[1].style = 'orange3'
    table.columns[1].header_style = 'bright_cyan'

    table.add_row('MD5', bytes_info.md5)
    table.add_row('SHA1', bytes_info.sha1)
    table.add_row('SHA256', bytes_info.sha256)
    return table



from dataclasses import dataclass, field

from rich.text import Text

from yaralyzer.encoding_detection.encoding_assessment import EncodingAssessment
from yaralyzer.util.helpers.rich_helper import na_txt

DECODE_NOT_ATTEMPTED_MSG = Text('(decode not attempted)', style='no_attempt')


@dataclass
class DecodingTableRow:
    encoding_label: Text
    confidence_text: Text
    errors_while_decoded: Text  # This is really "is_forced"?
    decoded_string: Text
    # Properties below here are not displayed in the table but are used for sorting etc.
    confidence: float
    encoding: str
    sort_score: float
    encoding_label_plain: str = field(init=False)

    def __post_init__(self):
        self.encoding_label_plain = self.encoding_label.plain

    def to_row_list(self) -> list[Text]:
        """Returns a row for the decoding attempts table."""
        return [self.encoding_label, self.confidence_text, self.errors_while_decoded, self.decoded_string]

    @classmethod
    def from_decoded_assessment(cls, assessment: EncodingAssessment, is_forced: Text, txt: Text, score: float) -> 'DecodingTableRow':
        """
        Alternate constructor that builds a table row for a decoding attempt.

        Args:
            assessment (EncodingAssessment): The `chardet` assessment for the encoding used.
            is_forced (Text): Text indicating if the decode was forced.
            txt (Text): The decoded string as a rich `Text` object (with highlighting).
            score (float): The score to use for sorting this row in the table.
        """
        return cls(
            encoding_label=assessment.encoding_label,
            confidence_text=assessment.confidence_text,
            errors_while_decoded=is_forced,
            decoded_string=txt,
            confidence=assessment.confidence,
            encoding=assessment.encoding,
            sort_score=score,
        )

    @classmethod
    def from_undecoded_assessment(cls, assessment: EncodingAssessment, score: float) -> 'DecodingTableRow':
        """
        Alternate constructor for a row with just `chardet` assessment confidence data and no actual
        decoding attempt string.

        Args:
            assessment (EncodingAssessment): The `chardet` assessment for the encoding used.
            score (float): The score to use for sorting this row within the table.
        """
        return cls.from_decoded_assessment(assessment, na_txt(), DECODE_NOT_ATTEMPTED_MSG, score)



"""
Helps with `chardet` library.
"""
from dataclasses import dataclass, field
from typing import Any, Optional

from chardet.resultdict import ResultDict
from rich.text import Text

from yaralyzer.encoding_detection.character_encodings import ENCODING
from yaralyzer.util.helpers.rich_helper import DIM_COUNTRY_THRESHOLD, meter_style, prefix_with_style

CONFIDENCE = 'confidence'
LANGUAGE = 'language'


@dataclass
class EncodingAssessment:
    """
    Class to smooth some of the rough edges around the `dict`s returned by `chardet.detect_all()`.

    Attributes:
        assessment (ResultDict): The dict returned by `chardet.detect_all()`.
        encoding (str): The encoding detected, in lowercase.
        confidence (float): Confidence score from 0.0 to 100.0.
        confidence_text (Text): Rich `Text` object representing the confidence with styling.
        language (Optional[str]): The detected language, if any.
        encoding_label (Text): Rich `Text` object for displaying the encoding with optional language info.
    """

    assessment: ResultDict
    confidence_text: Text = field(init=False)
    encoding_label: Text = field(init=False)

    @property
    def encoding(self) -> str:
        return self.assessment[ENCODING].lower()

    @property
    def confidence(self) -> float:
        """Shift confidence from 0-1.0 scale to 0-100.0 scale"""
        return 100.0 * (self._get_dict_empty_value_as_None(CONFIDENCE) or 0.0)

    @property
    def language(self) -> str | None:
        return self._get_dict_empty_value_as_None(LANGUAGE)

    def __post_init__(self) -> None:
        self.confidence_text = prefix_with_style(f"{round(self.confidence, 1)}%", style=meter_style(self.confidence))
        # Add detected language info and label if any language was detected
        self.set_encoding_label(self.language.title() if self.language else None)

    @classmethod
    def dummy_encoding_assessment(cls, encoding: str) -> 'EncodingAssessment':
        """
        Construct an empty `EncodingAssessment` to use as a dummy when `chardet` gives us nothing.

        Args:
            encoding (str): The encoding to use for the dummy assessment.
        """
        assessment = cls(ResultDict(encoding=encoding, confidence=0.0, language=None))
        assessment.confidence_text = Text('none', 'no_attempt')
        return assessment

    def set_encoding_label(self, alt_text: str | None) -> None:
        """
        Alt text is displayed below the encoding in slightly dimmer font.

        Args:
            alt_text (str | None): Text to display along with the encoding (often the inferred language)
        """
        self.encoding_label = Text(self.encoding, 'encoding.header')

        if alt_text is not None:
            dim = 'dim' if (self.confidence or 0.0) < DIM_COUNTRY_THRESHOLD else ''
            self.encoding_label.append(f" ({alt_text})", style=f"color(23) {dim}")

    def __rich__(self) -> Text:
        return Text('<Chardet(', 'white') + self.encoding_label + Text(':') + self.confidence_text + Text('>')

    def __str__(self) -> str:
        return self.__rich__().plain

    def _get_dict_empty_value_as_None(self, key: str) -> Any:
        """Return `None` if the value at `key` is an empty string, empty list, etc."""
        value = self.assessment.get(key)

        if isinstance(value, (dict, list, str)) and len(value) == 0:
            return None
        else:
            return value



"""
Color theme stuff. [Rich color names](https://rich.readthedocs.io/en/stable/appendix/colors.html)
TODO: interesting colors # row_styles[0] = 'reverse bold on color(144)' <-
"""
from rich.columns import Columns
from rich.console import Group
from rich.panel import Panel
from rich.padding import Padding
from rich.style import Style
from rich.terminal_theme import TerminalTheme
from rich.text import Text
from rich.theme import Theme

# Colors
ALERT_STYLE = 'error'  # Regex Capture used when extracting quoted chunks of bytes
BYTES = 'color(100) dim'
BYTES_NO_DIM = 'color(100)'
BYTES_BRIGHTEST = 'color(220)'
BYTES_BRIGHTER = 'orange1'
BYTES_HIGHLIGHT = 'color(136)'
DANGER_HEADER = 'color(88) on white'  # Red
DEFAULT_HIGHLIGHT_STYLE = 'orange1'
DARK_GREY = 'color(236)'
GREY = 'color(241)'
GREY_ADDRESS = 'color(238)'
PEACH = 'color(215)'

# For the table shown by running --show-colors
MAX_THEME_COL_SIZE = 35

# Theme used by main console
YARALYZER_THEME_DICT = {
    # colors
    'dark_orange': 'color(58)',
    'grey': GREY,
    'grey.dark': DARK_GREY,
    'grey.dark_italic': f"{DARK_GREY} italic",
    'grey.darker_italic': 'color(8) dim italic',
    'grey.darkest': 'color(235) dim',
    'grey.light': 'color(248)',
    'off_white': 'color(245)',
    'zero_bytes': 'color(20)',
    # data types
    'encoding': 'color(158) underline bold',
    'encoding.header': 'color(158) bold',
    'encoding.language': 'dark_green italic',
    'number': 'cyan',
    'regex': 'color(218) dim',
    'no_attempt': "color(60) dim italic",
    # design elements
    'decode.section_header': 'color(100) reverse',
    'decode.subheading': PEACH,
    'decode.subheading_2': 'color(215) dim italic',
    'decode.table_header': 'color(101) bold',
    'headline': 'bold white underline',
    # bytes
    'ascii': 'color(58)',
    'ascii_unprintable': 'color(131)',
    'bytes': BYTES,
    'bytes.title_dim': 'orange1 dim',
    'bytes.title': BYTES_BRIGHTER,
    'bytes.decoded': BYTES_BRIGHTEST,
    # yara
    'matched_rule': 'on bright_black bold',
    'yara.key': DARK_GREY,
    'yara.match_var': 'color(156) italic',
    'yara.string': 'white',
    'yara.date': 'color(216)',
    'yara.url': 'color(220)',
    'yara.int': 'color(45)',
    'yara.hex': 'color(98)',
    'yara.scanned': Style(color='yellow', underline=True, bold=True),
    'yara.rules':  Style(color='color(135)', underline=True, bold=True),
    # error log events
    'error': 'bright_red',
    # override defaults
    'repr.path': 'orchid2',
    'repr.filename': 'orchid2',
}

LOG_THEME_DICT = {
    'repr.path': 'dark_orange3',
    'repr.filename': 'dark_orange3',
    'repr.none': 'grey23 italic',
    **YARALYZER_THEME_DICT,
}

LOG_THEME = Theme(LOG_THEME_DICT)
YARALYZER_THEME = Theme(YARALYZER_THEME_DICT)

# Used by --env-vars option
CLI_OPTION_TYPE_STYLES = {
    'Dir': 'violet',
    'Path': 'magenta',
    'Pattern': 'orange1',
    'bool': 'bright_red',
    'int': 'cyan',
    'float': 'blue',
    'str': 'green',
}


# TerminalThemes are used when saving SVGS. This one just swaps white for black in DEFAULT_TERMINAL_THEME
YARALYZER_TERMINAL_THEME = TerminalTheme(
    (0, 0, 0),
    (255, 255, 255),
    [
        (0, 0, 0),
        (128, 0, 0),
        (0, 128, 0),
        (128, 128, 0),
        (0, 0, 128),
        (128, 0, 128),
        (0, 128, 128),
        (192, 192, 192),
    ],
    [
        (128, 128, 128),
        (255, 0, 0),
        (0, 255, 0),
        (255, 255, 0),
        (0, 0, 255),
        (255, 0, 255),
        (0, 255, 255),
        (255, 255, 255),
    ],
)


def color_theme_grid(styles: dict, app_name: str) -> Padding:
    """Lay out the colors in 'styles' in a grid with a header panel."""
    panel = Panel(f'The {app_name.title()} Color Theme', style='honeydew2', width=60)

    colors = [
        Text('', style='white').append(name[:MAX_THEME_COL_SIZE], style=style).append(' ')
        for name, style in styles.items()
        if name not in ['reset', 'repr_url']
    ]

    group = Group(panel, Text(''), Columns(colors, column_first=True, padding=(0, 5), equal=True))
    return Padding(group, (1, 2))


def theme_colors_with_prefix(prefix: str) -> list[Text]:
    """Return a list of (name, style) `Text` objects for all styles in the theme that start with `prefix`."""
    return [Text(k, v) for k, v in YARALYZER_THEME.styles.items() if k.startswith(prefix)]



class classproperty:
    """
    Decorator that mimics chaining @classmethod and @property for a getter. From:
    https://stackoverflow.com/questions/76249636/class-properties-in-python-3-11
    """
    def __init__(self, func):
        self.fget = func

    def __get__(self, instance, owner):
        return self.fget(owner)



from argparse import ArgumentTypeError
from rich.padding import Padding
from rich.text import Text

from yaralyzer.util.helpers.env_helper import is_invoked_by_pytest
from yaralyzer.util.logging import highlighter, log_console


class InvalidArgumentError(ArgumentTypeError):
    pass


def handle_argument_error(msg: str, e: Exception | None = None, is_standalone_mode: bool = False) -> None:
    """Standalone mode means in a situation where the `yaralyze` command is being run."""
    if is_standalone_mode and not is_invoked_by_pytest():
        print_fatal_error_and_exit(msg, e)
    else:
        raise e or InvalidArgumentError(msg)


def print_fatal_error(msg: str | Text | None, e: Exception | None = None) -> None:
    """
    Print a fatal error message

    Args:
        msg (str): The error message to display.
        e (Exception | None): The exception that caused the error, if any.
    """
    txt = Text('').append('(ERROR)', style='bright_red bold').append(" ")

    if msg:
        msg = msg if isinstance(msg, Text) else Text(msg, style='honeydew2')
        txt.append(highlighter(msg))

    if e:
        txt.append(f" (caused by {type(e).__name__}: {e})", style='dim')

    log_console.print(Padding(txt, (1, 0, 0, 0)))


def print_fatal_error_and_exit(msg: str, e: Exception | None = None, exit_code: int = 1) -> None:
    """Print an error message and exit with code 'exit_code'."""
    print_fatal_error(msg, e)
    exit(exit_code)



import functools
import signal

from yaralyzer.util.helpers.env_helper import is_windows


def timeout(seconds=5, default=None):
    """
    From: https://stackoverflow.com/questions/75928586/how-to-stop-the-execution-of-a-function-in-python-after-a-certain-time

    Example:
        @timeout(seconds=5, default=None)
        def function():
            sleep(6000)
    """
    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            def handle_timeout(signum, frame):
                raise TimeoutError(f"Timed out after {seconds} seconds.")

            # Windows doesn't support SIGALRM
            if not is_windows():
                signal.signal(signal.SIGALRM, handle_timeout)
                signal.alarm(seconds)

            result = func(*args, **kwargs)

            if not is_windows():
                signal.alarm(0)
            return result

        return wrapper

    return decorator



"""
Utility methods for running shell commands, including some code smelly pytest related stuff
that needs to be here because it's used by the pytest suite in this repo and in Pdfalyzzer.
"""
import logging
import re
import shutil
from dataclasses import dataclass, field
from os import environ, getlogin
from pathlib import Path
from subprocess import CalledProcessError, CompletedProcess, run
# from typing import Self  # TODO: this requires python 3.11

from yaralyzer.util.constants import INKSCAPE
from yaralyzer.util.helpers.env_helper import PYTEST_REBUILD_FIXTURES_ENV_VAR, _should_rebuild_fixtures
from yaralyzer.util.helpers.file_helper import load_file, relative_path
from yaralyzer.util.helpers.string_helper import strip_ansi_colors
from yaralyzer.util.logging import LOG_SEPARATOR, invocation_str, log, log_bigly, log_console
from yaralyzer.util.timeout import timeout

WROTE_TO_FILE_REGEX = re.compile(r"Wrote '(.*)' in [\d.]+ seconds")


@dataclass
class ShellResult:
    """
    Wrapper around `CompletedProcess` to help with some common tasks like stripping ansi color codes.
    Also used by pytest to compare output of commands to prerecorded fixture data.

    Attributes:
        result (CompletedProcess): Object returned by `subprocess.run()`
        no_log_args (list[str], optional): Arguments that we would prefer not to see in the logs.
    """
    result: CompletedProcess
    no_log_args: list[str] = field(default_factory=list)

    @property
    def invocation_str(self) -> str:
        """Simplified version of the command that was run."""
        return invocation_str([arg for arg in self.result.args if arg not in (self.no_log_args)])

    @property
    def stderr(self) -> str:
        return self.result.stderr.decode() if isinstance(self.result.stderr, bytes) else self.result.stderr

    @property
    def stderr_stripped(self) -> str:
        return strip_ansi_colors(self.stderr)

    @property
    def stdout(self) -> str:
        return self.result.stdout.decode() if isinstance(self.result.stdout, bytes) else self.result.stdout

    @property
    def stdout_lines(self) -> list[str]:
        return self.stdout_stripped.split('\n')

    @property
    def stdout_stripped(self) -> str:
        return strip_ansi_colors(self.stdout)

    @classmethod
    def from_cmd(cls, cmd: str | list, verify_success: bool = False, no_log_args: list[str] | None = None) -> 'ShellResult':
        """
        Alternate constructor that runs `cmd` and gets the result.

        Args:
            cmd (str | list): The shell command to run.
            verify_success (bool, optional): If True run `check_returncode()` (raises on non-zero return codes).
            no_log_args (list[str], optional): Args that might be in `cmd` to not show in logs.
        """
        result = cls(run(safe_args(cmd), capture_output=True, env=environ, text=True), no_log_args or [])
        log.debug(f"Ran command: {result.invocation_str}")

        if verify_success:
            try:
                result.result.check_returncode()
            except (CalledProcessError, FileNotFoundError) as e:
                log_bigly(f"Shell command returned error code! ({e})", result.output_logs(), logging.ERROR)
                raise e

        return result

    def compare_exported_files_to_existing(self, against_dir: Path, only_last_file: bool = False) -> None:
        """
        Compare the file exported by this command to an existing file in `against_dir`
        This command should have written to the same filename just in a different dir.

        Args:
            against_dir (Path): Dir where the file to compare against exists already.
            only_last_file (bool, optional): If True only compare the last file that was written.
        """
        exported_paths = self.exported_file_paths()[-1:] if only_last_file else self.exported_file_paths()

        for exported_path in exported_paths:
            assert exported_path.exists(), f"'{exported_path}' does not exist, {self.output_logs()}"
            existing_path = relative_path(against_dir.joinpath(exported_path.name))
            exported_data = load_file(exported_path)

            if _should_rebuild_fixtures():
                if getlogin() in strip_ansi_colors(exported_data):
                    raise ValueError(f"Found local username in exported fixture data!")

                log.warning(f"\nOverwriting fixture '{existing_path}'\n   with contents of '{exported_path}'")
                shutil.move(exported_path, existing_path)
                continue

            assert existing_path.exists(), f"Existing file we want to compare against '{existing_path}' doesn't exist!"
            existing_data = load_file(existing_path)

            # # Sometimes pytests diff is very, very slow, so we put a timer on it and fall back to showing diff cmd.
            @timeout(seconds=5)
            def compare_files():
                assert exported_data == existing_data, self._fixture_mismatch_log_msg(existing_path, exported_path)

            try:
                compare_files()
            except TimeoutError:
                log.error(f"Got timeout error when comparing '{existing_path}' to '{exported_path}'!")
                is_output_same_as_fixture = exported_data == existing_data
                assert is_output_same_as_fixture, self._fixture_mismatch_log_msg(existing_path, exported_path)

            log.debug(f"Validated '{exported_path}' as matching the exiting file...")

    def exported_file_paths(self) -> list[Path]:
        """Finds the last match."""
        written_paths = [relative_path(Path(m.group(1))) for m in WROTE_TO_FILE_REGEX.finditer(self.stderr_stripped)]
        assert len(written_paths) > 0, f"Could not find 'wrote to file' msg in stderr:\n\n{self.stderr}"
        log_bigly(f"Found {len(written_paths)} written files in the logs", self.output_logs())
        return written_paths

    def last_exported_file_path(self) -> Path:
        """Returns the last file that exported by this shell command."""
        return self.exported_file_paths()[-1]

    def output_logs(self, with_streams: bool = False) -> str:
        """Long string with all info about a shell command's execution and output."""
        cmd = invocation_str([arg for arg in self.result.args if arg not in (self.no_log_args or [])])
        msg = f"Return code {self.result.returncode} from shell command:\n\n{cmd}"

        if with_streams:
            for i, stream in enumerate([self.stdout, self.stderr]):
                label = 'stdout' if i == 0 else 'stderr'
                msg += f"\n\n\n\n[{label}]\n{LOG_SEPARATOR}\n{stream}\n{LOG_SEPARATOR}"

        return msg + "\n"

    def _fixture_mismatch_log_msg(self, existing_path: Path, export_path: Path) -> str:
        error_msg = f"Contents of '{export_path}' does not match fixture: '{existing_path}'\n\n"
        error_msg += f"Fixtures can be updated by running '{PYTEST_REBUILD_FIXTURES_ENV_VAR}=True pytest tests/test_file_export.py'\n\n"  # noqa: E501
        error_msg += f"pytest diffs can be slow, here's the manual diff cmd:\n\n   diff '{existing_path}' '{export_path}'\n\n"  # noqa: E501
        error_msg += f"Result of diff:\n\n"

        try:
            diff_output = type(self).from_cmd(['diff', '-a', existing_path, export_path])
            print(f"Result of diff '{existing_path}'\n       against '{export_path}'\n\n")
            print(diff_output.output_logs(True))
            print(f"\n\n\nRaw diff stdout:\n\n", diff_output.stdout)
        except Exception as e:
            log_console.print(f"Failed to print diff of '{existing_path}'\n        against '{export_path}'!", style='bright_red bold')

        return error_msg

    @classmethod
    def run_and_compare_exported_files_to_existing(
        cls,
        cmd: list[str] | str,
        against_dir: Path,
        no_log_args: list[str] | None = None,
    ) -> 'ShellResult':
        """
        Used by pytest to compare fixture data to output of export commands. Here so that pdfalyzer can use it.

        Args:
            cmd (list[str] | str): Shell command to run.
            against_dir (Path): Dir where the existing files you want to compare the new oones to live.
            ignorable_args (list[str], optional): Don't log these args if they exist in `cmd`.
        """
        shell_result = cls.from_cmd(cmd, True, no_log_args)
        shell_result.compare_exported_files_to_existing(against_dir)
        return shell_result


def get_inkscape_version() -> str | None:
    """Check to see if Inkscape is installed on the current sytem and if so find its version number."""
    try:
        result = ShellResult.from_cmd([INKSCAPE, '--version'], verify_success=True)
        return result.stdout.lower().removeprefix(INKSCAPE).split()[0] or '(unknown)'
    except (CalledProcessError, FileNotFoundError):
        pass


def safe_args(cmd: str | list) -> list[str]:
    """Make sure everything is a string and not, for instance, a `Path`."""
    args = cmd.split() if isinstance(cmd, str) else cmd
    return [str(arg) for arg in args]



"""
Holds the rich.Console instance that controls the stdout printing and file export.
"""
from os import devnull
from sys import argv

from rich.console import Console
from rich.errors import MarkupError
from rich.style import Style
from rich.text import Text

from yaralyzer.util.constants import SUPPRESS_OUTPUT_OPTION
from yaralyzer.util.helpers.env_helper import default_console_kwargs, stderr_console
from yaralyzer.output.theme import YARALYZER_THEME, color_theme_grid


console_kwargs = default_console_kwargs()

if SUPPRESS_OUTPUT_OPTION in argv:
    stderr_console.print(f"Suppressing terminal output because {SUPPRESS_OUTPUT_OPTION} is enabled...", style='dim')
    console_kwargs.update({'file': open(devnull, "wt")})

# This is the global stdout manager
console = Console(highlight=False, theme=YARALYZER_THEME, **console_kwargs)


def console_print_with_fallback(_string: Text | str, style=None) -> None:
    """`rich.console.print()` with fallback to regular `print()` if there's a Rich Markup issue."""
    try:
        console.print(_string, style=style)
    except MarkupError:
        console.print(f"Hit a bracket issue with rich.console printing, defaulting to plain print", style='warn')
        print(_string.plain if isinstance(_string, Text) else _string)


def console_width() -> int:
    """Current width set in `console` object."""
    return console._width or 80  # TODO: wtf?


def show_color_theme(theme_dict: dict[str, str | Style], app_name: str) -> None:
    """Script method to show yaralyzer's color theme. Invocable with 'yaralyzer_show_colors'."""
    console.print(color_theme_grid(theme_dict, app_name))



"""
Configuration management for Yaralyzer.
"""
import os
import platform
import re
import sys
from contextlib import contextmanager
from copy import deepcopy
from os import environ
from pathlib import Path
from shutil import get_terminal_size
from typing import Any, Generator, Literal, Mapping, Sequence

from dotenv import load_dotenv
from rich.console import Console

from yaralyzer.util.constants import INVOKED_BY_PYTEST, YARALYZER_UPPER, dotfile_name
from yaralyzer.util.helpers.file_helper import relative_path

DEFAULT_CONSOLE_WIDTH = 160
PATH_ENV_VAR_REGEX = re.compile(r"^.*_(DIR|FILE|PATH)S?$", re.I)
PYTEST_REBUILD_FIXTURES_ENV_VAR = 'PYTEST_REBUILD_FIXTURES'
DOTFILE_DIRS = [Path.cwd(), Path.home()]

is_linux = lambda: platform.system().lower() == 'linux'
is_macos = lambda: platform.system().lower() == 'darwin'
is_windows = lambda: platform.system().lower() == 'windows'


def config_var_name(env_var: str) -> str:
    """
    Get the name of `env_var` and strip off `YARALYZER_` prefix.

    Example:
        ```
        config_var_name('YARALYZER_SURROUNDING_BYTES') => 'SURROUNDING_BYTES'
        ```
    """
    env_var = env_var.removeprefix(f"{YARALYZER_UPPER}_")
    return f'{env_var=}'.partition('=')[0]


# TODO: why is this a function?
def console_width_possibilities():
    """Returns a list of possible console widths, the first being the current terminal width."""
    # Subtract 2 from terminal cols just as a precaution in case things get weird
    return [get_terminal_size().columns - 2, DEFAULT_CONSOLE_WIDTH]


def is_cairosvg_installed() -> bool:
    """True if cairosvg package is available on the current system."""
    try:
        import cairosvg
        return True
    except (ModuleNotFoundError, OSError):
        return False


def is_env_var_set_and_not_false(var_name: str) -> bool:
    """Return `True` if `var_name` is not empty and set to anything other than "false" (capitalization agnostic)."""
    if var_name in environ:
        var_value = environ[var_name]
        return var_value is not None and len(var_value) > 0 and var_value.lower() != 'false'
    else:
        return False


def is_github_workflow() -> bool:
    return is_env_var_set_and_not_false('GITHUB_ACTION')


def is_invoked_by_pytest() -> bool:
    """Return `True` if invoked in a `pytest` context."""
    return is_env_var_set_and_not_false(INVOKED_BY_PYTEST)


def is_path_var(env_var_name: str) -> bool:
    """Returns True if `env_var_name` ends with _DIR or _PATH."""
    return bool(PATH_ENV_VAR_REGEX.match(env_var_name))


def load_dotenv_file(app_name: Literal['pdfalyzer', 'yaralyzer']) -> None:
    if is_invoked_by_pytest():
        return

    for dotenv_file in [dir.joinpath(dotfile_name(app_name)) for dir in DOTFILE_DIRS]:
        if dotenv_file.exists():
            load_dotenv(dotenv_path=dotenv_file)
            lines = [l for l in dotenv_file.read_text().split('\n') if not l.startswith('#')]
            stderr_console.print(f"Loaded {len(lines)} vars from {relative_path(dotenv_file)}...", style='dim')
            return


@contextmanager
def temporary_argv(new_argv: Sequence[str | Path]) -> Generator[Any, Any, Any]:
    """Temporarily replace sys.argv with something else."""
    old_argv = list(sys.argv)
    sys.argv = [str(arg) for arg in new_argv]

    try:
        yield
    finally:
        sys.argv = old_argv


@contextmanager
def temporary_env(env_vars: Mapping[str, str | Path]) -> Generator[Any, Any, Any]:
    """
    Temporarily add variables to the environemnt.
    See: https://shay-palachy.medium.com/temp-environment-variables-for-pytest-7253230bd777

    Example:
        with temporary_env({'new_var': 'new_value}):
            do_stuff()
    """
    old_environ = dict(environ)
    environ.update({k: str(v) for k, v in env_vars.items()})

    try:
        yield
    finally:
        environ.clear()
        environ.update(old_environ)


def _should_rebuild_fixtures() -> bool:
    """
    True if pytest should overwrite fixture data with new output instead of comparing.
    It's here so that pdfalyzer can also use it.
    """
    return is_env_var_set_and_not_false(PYTEST_REBUILD_FIXTURES_ENV_VAR)


# Maximize output width if YARALYZER_MAXIMIZE_WIDTH is set (also can changed with --maximize-width option)
if is_invoked_by_pytest():
    CONSOLE_WIDTH = DEFAULT_CONSOLE_WIDTH
elif is_env_var_set_and_not_false('YARALYZER_MAXIMIZE_WIDTH'):
    CONSOLE_WIDTH = max(console_width_possibilities())
else:
    CONSOLE_WIDTH = min(console_width_possibilities())

DEFAULT_CONSOLE_KWARGS = {
    'color_system': 'auto',
    'width': CONSOLE_WIDTH,
}


def default_console_kwargs() -> dict[str, str | int]:
    """Returns a fresh copy of DEFAULT_CONSOLE_KWARGS."""
    return deepcopy(DEFAULT_CONSOLE_KWARGS)


# For use when you need to write output before the main rich.console has managed to get set up.
stderr_console = Console(stderr=True, **{**DEFAULT_CONSOLE_KWARGS, 'width': max(console_width_possibilities())})
# stderr_console.print(f"\n\n MAX WIDTH = {max(console_width_possibilities())}", style='bright_cyan')
# print(f"color_system: {stderr_console.color_system}")
